from cflib.crazyflie import Crazyflie

from numpy import zeros, eye



class CrazyDragon( Crazyflie ):

    def __init__( self ):
        super().__init__( rw_cache='./cache' )

        self.pos         = zeros(3)
        self.vel         = zeros(3)
        self.att         = zeros(3)
        self.acc         = zeros(3)
        self.command     = zeros(3)
        self.destination = zeros(3)
        self.rot         = eye(3)

        self.ready_for_command = False



from ..crazy import CrazyDragon

from threading import Thread

from .._base._navigation_base.imu       import IMU
from .._base._navigation_base.imu_setup import preflight_sequence
from .._base._navigation_base.qualisys  import Qualisys

from time import sleep



class Navigation( Thread ):

    def __init__( self, cf: CrazyDragon, config ):

        super().__init__()

        self.daemon = True

        self.cf = cf

        self.imu = IMU( cf )
        self.qtm = Qualisys( config['body_name'] )

        self.navigate = True


    @classmethod
    def _on_pose( cls, cf: CrazyDragon, data: list ):
        
        cf.pos[:] = data[0:3]
        cf.att[:] = data[3:6]

        cf.extpos.send_extpos( data[0], data[1], data[2] )


    def run( self ):

        cf = self.cf

        imu = self.imu
        qtm = self.qtm

        preflight_sequence( cf )

        sleep( 1 )

        imu.start_get_acc()
        imu.start_get_vel()

        qtm.on_pose = lambda pose: __class__._on_pose( cf, pose )

        while self.navigate:

            sleep( 0.1 )


    def join( self ):

        self.navigate = False

        super().join()


from threading import Thread

import xml.etree.cElementTree as ET

import qtm
import asyncio

from math import isnan



class Qualisys( Thread ):

    def __init__( self, _cfs: dict ):

        Thread.__init__( self )

        if ( type( _cfs ) != dict ):
            self._cfs = { _cfs: None }
        else:
            self._cfs = _cfs

        self.on_pose    = {}
        self.connection = None
        self.qtm_6DoF_labels = []
        self._stay_open = True

        self.start()


    def close( self ):

        self._stay_opne = False

        self.join()


    def run( self ):

        asyncio.run( self._life_cycle() )


    async def _life_cycle( self ):

        await self._connect()

        while ( self._stay_open ):

            await asyncio.sleep( 1 )

        await self._close()


    async def _connect( self ):

        self.connection = await qtm.connect( '127.0.0.1', version='1.23' )

        if self.connection is None:
            print( "Failed to connect" )
            return

        params      = await self.connection.get_parameters( parameters=['6d'] )
        self.params = params

        xml = ET.fromstring( params )

        self.qtm_6DoF_labels = [label.text.strip() for index, label in enumerate(xml.findall( '*/Body/Name' ) ) ]

        await self.connection.stream_frames(
            components=[ '6deuler' ],
            on_packet =self._on_packet
        )

    
    async def _close( self ):

        await self.connection.stream_frames_stop()
        self.connection.disconnect()


    async def _discover( self ):

        async for qtm_instance in qtm.Discover( '192.168.254.1' ):
            return qtm_instance


    def _on_packet( self, packet ):

        header, bodies  = packet.get_6d_euler()

        for bodyname, _ in self._cfs.items():

            if bodyname not in self.qtm_6DoF_labels:
                print( 'Body' + bodyname + 'not found.' )
            else:
                index = self.qtm_6DoF_labels.index( bodyname )

                if bodies is None:
                    pass
            
                else:
                    data = bodies[index]

                    position = data[0]
                    x = position[0] / 1000
                    y = position[1] / 1000
                    z = position[2] / 1000

                    euler = data[1]
                    R = euler[2]
                    P = euler[1]
                    Y = euler[0]

                    if self.on_pose:
                        if isnan( x ):
                            return

                        self.on_pose( [x, y, z, R, P, Y] )


## empty

from .constants import g, alpha, Kp

from .integral_loop import _dot_thrust
from .integral_loop import _thrust_clip

from .optimus_prime import _command_as_RPY
from .optimus_prime import _command_is_not_in_there



from ...crazy import CrazyDragon

from .smoother import smooth_command

from .constants import Kp, Kd, g

from numpy        import array, zeros
from numpy.linalg import norm

from time import sleep



def landing( cf: CrazyDragon, option=1, h=0.2, T=5, dt=0.1, step=0.075 ):

    cur     = zeros(3)
    des     = zeros(3)
    des_cmd = zeros(3)
    acc_cmd = zeros(3)
    P_pos   = zeros(3)
    D_pos   = zeros(3)
    care_g  = array([0,0,g])

    print( 'landing' )

    n = int( T / dt )
    t = 0

    command = cf.command

    cur[:] = cf.pos
    pos    = cf.pos
    vel    = cf.vel

    if option:
        des[0] = cur[0]
        des[1] = cur[1]
        des[2] = h
    else:
        des[0] = 0
        des[1] = 0
        des[2] = h

    cf.destination[:] = des

    for _ in range( n ):

        des_cmd[:] = smooth_command( 
            des, cur, t, 5.0
        )

        P_pos[:] = des_cmd - pos
        D_pos[:] = vel

        if ( norm( pos - des ) ) < 0.2:
            break

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] -= D_pos * Kd
        acc_cmd[:] += care_g

        command[:] = acc_cmd

        t += dt

        sleep( dt )


from ...crazy import CrazyDragon

from ...recorder import Recorder

# from .constants import Kp, Kd, g

from numpy        import array, zeros
from numpy.linalg import norm

from time import sleep



w = array([0.700,0.700,0.700])
j = array([1.600,1.600,0.700])

Kp = w * w
Kd = 2 * j * w


def landing_supporter( cf: CrazyDragon, recorder:Recorder, option=1, dt=0.1, step=0.03 ):

    des     = zeros(3)
    acc_cmd = zeros(3)
    P_pos   = zeros(3)
    D_pos   = zeros(3)
    care_g  = array([0,0,9.81])

    print( 'landing supporter', recorder.record_length )

    pos = cf.pos
    vel = cf.vel

    des[:] = pos

    for _ in range( 30 ):

        if ( des[2] > 0 ):
            des[:] -= step
        else:
            des[:] = 0

        if ( norm( pos ) < 0.02 ):
            break

        P_pos[:] = des - pos
        D_pos[:] = vel

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] -= D_pos * Kd
        acc_cmd[:] += care_g

        cf.command[:] = acc_cmd

        sleep( dt )

    acc_cmd[:] = care_g

    for _ in range( 10 ):

        acc_cmd[2] -= step

        cf.command[:] = acc_cmd

    print( 'land' )


import matplotlib.pyplot as plt



def plot_acc_pos_cmd( acc, acccmd, vel, pos, posref, _len, _stt, _end, date ):

    fig = plt.figure( figsize=(14,14) )

    ax1 = fig.add_subplot( 331 )
    ax1.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax1.plot(acc[0,:_len], label='realtime acceleration x')
    ax1.plot(acccmd[0,:_len], label='command input x')
    ax1.legend()
    ax1.grid()

    ax2 = fig.add_subplot( 332 )
    ax2.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax2.plot(acc[1,:_len], label='realtime acceleration y')
    ax2.plot(acccmd[1,:_len], label='command input y')
    ax2.legend()
    ax2.grid()

    ax3 = fig.add_subplot( 333 )
    ax3.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax3.plot(acc[2,:_len], label='realtime acceleration z')
    ax3.plot(acccmd[2,:_len], label='command input z')
    ax3.legend()
    ax3.grid()

    ax4 = fig.add_subplot( 334 )
    ax4.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax4.plot(vel[0,:_len], label='realtime velocity x')
    ax4.legend()
    ax4.grid()

    ax5 = fig.add_subplot( 335 )
    ax5.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax5.plot(vel[1,:_len], label='realtime velocity y')
    ax5.legend()
    ax5.grid()

    ax6 = fig.add_subplot( 336 )
    ax6.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax6.plot(vel[2,:_len], label='realtime velocity z')
    ax6.legend()
    ax6.grid()

    ax7 = fig.add_subplot( 337 )
    ax7.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax7.plot(pos[0,:_len], label='realtime position x')
    ax7.plot(posref[0,:_len], label='destination position x')
    ax7.legend()
    ax7.grid()

    ax8 = fig.add_subplot( 338 )
    ax8.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax8.plot(pos[1,:_len], label='realtime position y')
    ax8.plot(posref[1,:_len], label='destination position y')
    ax8.legend()
    ax8.grid()

    ax9 = fig.add_subplot( 339 )
    ax9.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax9.plot(pos[2,:_len], label='realtime position z')
    ax9.plot(posref[2,:_len], label='destination position z')
    ax9.legend()
    ax9.grid()

    plt.savefig( f'./flight_data/{date}/trg.png' )


def plot_thrust( thrust, thrustcmd, _len, _stt, _end, date ):

    fig = plt.figure( figsize=(8,8) )

    ax = fig.add_subplot()
    ax.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )

    ax.plot(thrust[:_len], label='realtime thrust')
    ax.plot(thrustcmd[:_len], label='reference thrust')
    ax.legend()
    ax.grid()

    plt.savefig( f'./flight_data/{date}/thr.png' )


def plot_att( att, attcmd, _len, _stt, _end, date ):

    fig = plt.figure( figsize=(14,14) )

    ax1 = fig.add_subplot( 311 )
    ax1.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax1.plot(att[0,:_len], label='realtime euler x')
    ax1.plot(attcmd[0,:_len], label='realtime euler input x')
    ax1.legend()
    ax1.grid()

    ax2 = fig.add_subplot( 312 )
    ax2.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax2.plot(att[1,:_len], label='realtime euler y')
    ax2.plot(attcmd[1,:_len], label='realtime euler input y')
    ax2.legend()
    ax2.grid()

    ax3 = fig.add_subplot( 313 )
    ax3.axvspan( _stt, _end, facecolor='gray', alpha=0.5 )
    ax3.plot(att[2,:_len], label='realtime euler z')
    ax3.plot(attcmd[2,:_len], label='realtime euler input z')
    ax3.legend()
    ax3.grid()

    plt.savefig( f'./flight_data/{date}/att.png' )


def plot_trj( pos, _len, _stt, _end, date ):

    fig = plt.figure( figsize=(14,14) )

    ax = fig.add_subplot( 111, projection='3d' )
    ax.plot( pos[0,    :_len], pos[1,_stt:_end], pos[2,_stt:_end], label='trajectory', linewidth=10, alpha=0.5 )
    ax.plot( pos[0,_stt:_end], pos[1,_stt:_end], pos[2,_stt:_end], label='optimal guidance' )

    ax.legend()

    plt.savefig( f'./flight_data/{date}/3d-trj.png' )


from threading import Thread

from ..swarm_handler import SwarmHandler

from ..._base._controller_base.integral_loop import _dot_thrust
from ..._base._controller_base.integral_loop import _thrust_clip

from ..._base._controller_base.optimus_prime import _command_as_RPY
from ..._base._controller_base.optimus_prime import _command_is_not_in_there

from ..._base._controller_base.constants import alpha

from ...crazy import CrazyDragon

from numpy import zeros, array
from numpy import zeros_like

from time import sleep




class SwarmController( Thread, SwarmHandler ):

    def __init__( self, _cfs: dict, config ):

        super().__init__()

        self.daemon = True

        self._cfs = _cfs

        self.dt = config['dt']
        self.n  = config['n']

        self.acc_cmd = zeros(3)
        self.command = zeros(4)
        self.thrust  = array( [alpha * 9.81], dtype=int )

    
    def init_send_setpoint( self ):

        _cfs = self._cfs

        self.parallel_run( self.__class__._init_send_setpoint, _cfs )

    
    @classmethod
    def _init_send_setpoint( cls, _cf: CrazyDragon ):
        
        commander = _cf.commander

        commander.send_setpoint( 0, 0, 0, 0 )

        _cf.ready_for_command = True


    def stop_send_setpoint( self ):

        _cfs = self._cfs

        self.parallel_run( self.__class__._stop_send_setpoint, _cfs )


    @classmethod
    def _stop_send_setpoint( cls, _cf: CrazyDragon ):

        commander = _cf.commander

        _cf.command[:] = 0

        _cf.ready_for_command = False

        for _ in range( 50 ):

            commander.send_setpoint( 0, 0, 0, 10001 )

            sleep( 0.02 )

        commander.send_stop_setpoint()


    def run( self ):
        
        _cfs = self._cfs

        n  = self.n
        dt = self.dt / n

        args_list = [ n, dt ]
        args_dict = {}

        for bodyname, _ in _cfs.items():
            args_dict[bodyname] = [ 
                zeros_like( self.acc_cmd ),
                zeros_like( self.command ),
                zeros_like( self.thrust )
            ]

        self.parallel_run( self.__class__._run, _cfs, args_list=args_list, args_dict=args_dict )


    @classmethod
    def _run( cls, _cf: CrazyDragon, *args ):

        commander = _cf.commander

        n  = args[0]
        dt = args[1]

        att_cur = _cf.att
        acc_cur = _cf.att

        acc_cmd = args[2]
        command = args[3]
        thrust  = args[4]
        
        while not _cf.ready_for_command:
            sleep( 0.1 )

        while _cf.ready_for_command:

            acc_cmd[:] = _cf.command

            _command_is_not_in_there( acc_cmd, att_cur )

            _command_as_RPY( acc_cmd, command )

            if ( acc_cmd[2] == 0 ):
                sleep( dt )
                return 
            
            for _ in range( n ):

                thrust[0] += _dot_thrust( command, acc_cur )

                thrust[0] = _thrust_clip( thrust[0] )

                commander.send_setpoint(
                    command[0],
                    command[1],
                    command[2],
                    thrust[0]
                )

                sleep( dt )


version = "__5.0.1__"

from .control import Controller

from .crazy import CrazyDragon

from .guidance import Guidance

from .guidance import utils

from .navigation import Navigation

from .recorder import Recorder



from numpy        import arcsin, rad2deg
from numpy        import sqrt
from numpy        import cos, sin, deg2rad

from numpy.linalg import norm



def _command_is_not_in_there( euler, acc_cmd ):
    ## yaw position
    yaw = euler[2]              ## [deg]
    yaw = deg2rad( yaw )        ## [rad]

    ## basic
    cy, sy = cos( yaw ), sin( yaw )

    ## update
    aE = cy * acc_cmd[0] + sy * acc_cmd[1]
    aN = cy * acc_cmd[1] - sy * acc_cmd[0]

    ## store
    acc_cmd[0] = aE
    acc_cmd[1] = aN

    return acc_cmd


def _command_as_RPY( acc_cmd, command ):
    """
    this function translates acc_cmd in ENU coordinate,
    to drone's body coordinate roll, pitch, yaw, acc

    coordinate rotates in 3 -> 2 -> 1 order

    command:
    1. roll
    2. pitch
    3. yaw
    4. acc
    """
    ## unpack acceleration command
    aE, aN, aU = acc_cmd        ## ENU coordinate

    ## acceleration to z-direction in drone's coordinate
    ## 3. yaw
    command[2] = 0              ## do not rotate, [deg/s]

    ## 2. pitch
    acc_str = sqrt( aE**2 + aU**2 )
    if acc_str:
        pitch_in_rad = arcsin( aE / acc_str )       ## [rad]
    else:
        pitch_in_rad = 0
    command[1] = rad2deg( pitch_in_rad )            ## [deg]

    ## 1. roll
    acc_str = sqrt( aN**2 + acc_str**2 )
    if acc_str:
        roll_in_rad = arcsin( -aN / acc_str )       ## [rad]
    else:
        roll_in_rad = 0
    command[0] = rad2deg( roll_in_rad )             ## [deg]

    ## acc strength
    command[3] = norm( acc_cmd )                    ## m/s^2


from ..crazy import CrazyDragon

from threading import Thread

from os import system

from datetime import datetime

from numpy import zeros, array

from time  import sleep

from ..control import alpha

from .visualizer import *



class Recorder( Thread ):


    def __init__(self, cf: CrazyDragon, commander, n=10000):
        ## for threading
        super().__init__()

        self.record_length = 0
        self.recording     = True

        self.guidance_start_idx = -1
        self.guidance_end_idx   = -1

        self.cf = cf

        ## callback functions
        self.record_callback = {
            'acccmd': array_type_data_callback,
            'vel'   : array_type_data_callback,
            'pos'   : array_type_data_callback,
            'posref': array_type_data_callback,
            'att'   : array_type_data_callback,
            'cmd'   : array_type_data_callback,
            'thrust': float_type_data_callback
        }
        ## data storage
        self.record_datastrg = {
            'acc'   : zeros((3,n)),
            'acccmd': zeros((3,n)),
            'vel'   : zeros((3,n)),
            'pos'   : zeros((3,n)),
            'posref': zeros((3,n)),
            'att'   : zeros((3,n)),
            'cmd'   : zeros((4,n)),
            'thrust': zeros((1,n))
        }
        ## realtime data
        self.realtime_data = {
            'acc'   : cf.acc,
            'acccmd': cf.command,
            'vel'   : cf.vel,
            'pos'   : cf.pos,
            'posref': cf.destination,
            'att'   : cf.att,
            'cmd'   : commander.command,
            'thrust': commander.thrust
        }

    
    def run(self):

        sleep(0.1)

        cf = self.cf

        while self.recording:

            self.record_datastrg['acc'][:,self.record_length] = self.realtime_data['acc']

            for key, callback in self.record_callback.items():

                datastrg = self.record_datastrg[key]
                data     = self.realtime_data[key]

                callback( datastrg, data, self.record_length )

            self.record_length += 1

            sleep(0.05)

    
    def stop_record(self):

        self.recording = False

    
    def guidance_start(self):

        self.guidance_start_idx = self.record_length


    def guidance_end(self):

        self.guidance_end_idx = self.record_length

    
    def join(self):

        self.recording = False

        super().join()

        _len = self.record_length
        _stt = self.guidance_start_idx
        _end = self.guidance_end_idx

        acc    = self.record_datastrg['acc']
        acccmd = self.record_datastrg['acccmd']

        vel    = self.record_datastrg['vel']

        pos    = self.record_datastrg['pos']
        posref = self.record_datastrg['posref']

        att    = self.record_datastrg['att']
        cmd    = self.record_datastrg['cmd']
        thrust = self.record_datastrg['thrust']

        d = datetime.now()
        date      = f'{d.year}-{d.month}-{d.day:02}-{d.hour:02}-{d.minute:02}-{d.second:02}'
        self.date = date

        system( f'cd ./flight_data && mkdir {date}' )

        plot_acc_pos_cmd( acc, acccmd, vel, pos, posref, _len, _stt, _end, date )
        plot_thrust( thrust[0,:], cmd[3,:]*alpha, _len, _stt, _end, date )
        plot_att( att, cmd[:3,:], _len, _stt, _end, date )

        plot_trj( pos, _len, _stt, _end, date )


def array_type_data_callback( datastrg, data, i ):

    datastrg[:,i] = array( data )


def float_type_data_callback( datastrg, data, i ):

    datastrg[0,i] = data